using System;
using System.Collections.Generic;
using System.Text;
using SharpNeatLib.NeuralNetwork;
using Engine;
using SharpNeatLib.NeatGenome;
using System.Xml;
using SharpNeatLib.CPPNs;

namespace Engine
{
    //Abstracts away neural networks and heterogeneity
    public class AgentBrain
    {
        //How many times should the network get activated. TODO make configurable
        private static int NET_ACTIVATION_STEPS = 2;

        //The "big" brain for heterogenous teams
        public INetwork brain;

        //Brains for homogenous teams
        private List<INetwork> brains; 

        //Homogenous team?
        public bool homogenous;
        public int numRobots;
        private List<Robot> robotListeners;
        private bool allActivated;
      
        private float[] teamInput;
        private bool[] activated;

        private SubstrateDescription substrateDescription;
        private INetwork genome;
        public NeatGenome ANN;

        private bool normalizeANNWeights;
        private bool adaptableANN;
        private bool modulatoryANN;

        public List<float> zcoordinates;

        public AgentBrain(bool homogenous, int numAgents, SubstrateDescription substrateDescription, INetwork genome,
            bool normalizeANNWeights, bool adaptableANN, bool modulatoryANN)
        {
            this.normalizeANNWeights = normalizeANNWeights;
            this.adaptableANN = adaptableANN;
            this.modulatoryANN = modulatoryANN;
            this.genome = genome;
            this.substrateDescription = substrateDescription;
            this.numRobots = numAgents;
            this.homogenous = homogenous;

            //inputCounter = 0;
            teamInput = new float[numAgents * substrateDescription.InputCount];
            activated = new bool[numAgents];

            createBrains();
            
            robotListeners = new List<Robot>();
        }

        private void createBrains()
        {
            if (genome != null)
            {
                if (homogenous)
                {

                    ANN = substrateDescription.generateHomogeneousGenome(genome, normalizeANNWeights);
                    brains = new List<INetwork>();
                    for (int i = 0; i < numRobots; i++)
                    {
                        INetwork b = ANN.Decode(null);
                        brains.Add(b);
                    }
                }
                else
                {

                    ANN = substrateDescription.generateMultiGenomeStack(genome, Convert.ToUInt32(numRobots), normalizeANNWeights, adaptableANN,
                                                                    modulatoryANN, out zcoordinates);
                    brain = ANN.Decode(null);
                }
            }
        }

        //Needs to be called if the number of sensors changes
        //Right now all brains have the same number of inputs
        public void updateInputDensity()
        {
            teamInput = new float[numRobots * substrateDescription.InputCount];
            //Check if the number of sensors changed so we have to regenerate the ANNs
            if (homogenous)
            {
                if (brains!=null && brains[0]!=null && brains[0].InputNeuronCount != substrateDescription.InputCount)
                {
                    Console.WriteLine("Recreating ANNs");
                    createBrains();
                }
            }
            else
            {
                if (brain!=null && (brain.InputNeuronCount/numRobots) != substrateDescription.InputCount)
                {
                    Console.WriteLine("Recreating ANNs");
                    createBrains();
                }

            }
        }

        //Robots need to be registered before they can receive ANN results
        public void registerRobot(Robot robot)
        {
            if (robotListeners.Contains(robot))
            {
                Console.WriteLine("Robot " + robot.id + " already registered");
                return;
            }
            robotListeners.Add(robot);
            if (robotListeners.Count > numRobots)
            {
                Console.WriteLine("Number of registered agents and number of agents does not match");
            }
        }

        public INetwork getBrain(int number)
        {
            if (homogenous)
            {
                return brains[number];
            }
            else
                return brain;   //only one brain for heterogenous teams
        }

        //Clear the listener list
        public void reset()
        {
            robotListeners.Clear();
        }

        //Activate the agents network with the given inputs
        public void setInputSignals(int agentNumber, float[] inputs)
        {
            if (homogenous)
            {
                brains[agentNumber].SetInputSignals(inputs);
                brains[agentNumber].MultipleSteps(NET_ACTIVATION_STEPS); //maybe base this on ANN depth

                float[] outputs = new float[brains[agentNumber].OutputNeuronCount];
                for (int j = 0; j < outputs.Length; j++)
                {
                    outputs[j] = brains[agentNumber].GetOutputSignal(j);
                }

                robotListeners[agentNumber].networkResults(outputs);
            
                return;
            }

            if (brain == null)
                return;

          //  Console.WriteLine("Received :"+agentNumber+" input length "+inputs.Length+" inputcounter "+inputCounter);

            for (int i = 0; i < inputs.Length; i++)
            {
                //if (inputCounter != i+agentNumber * inputs.Length)
                //{
                //    Console.WriteLine(inputCounter + " " + (i + agentNumber * inputs.Length));
                //}

                teamInput[(i + agentNumber * inputs.Length)] = inputs[i];
                //inputCounter++;
            }

            activated[agentNumber] = true;

            allActivated = true;
            for (int i = 0; i < activated.Length; i++)
            {
                if (!activated[i])
                {
                    allActivated = false;
                    break;
                }
            }

            if (allActivated)       //weight for all ANN to get activated before sending out the networkResult event
            {
   //             Console.WriteLine("All agents activated");
                allActivated = false;
             //   inputCounter = 0;
                for (int i=0;i<activated.Length; i++)
                {
                    activated[i]=false;
                }

                brain.SetInputSignals(teamInput); 
                brain.MultipleSteps(NET_ACTIVATION_STEPS); 

                int out_count = 0;
                int numOutputAgent = brain.OutputNeuronCount / numRobots;
                float[] outputs = new float[numOutputAgent];
                foreach (Robot robot in robotListeners)
                {
                    for (int y = 0; y < numOutputAgent; y++)
                    {
                        outputs[out_count % numOutputAgent] = brain.GetOutputSignal(out_count);

                        if (Double.IsNaN(outputs[out_count % numOutputAgent]))
                            Console.WriteLine("NaN in outputs");
                        out_count++;
                    }

      //              Console.WriteLine("Sending results to " + robot.id);
                    robot.networkResults(outputs); 
                }
            }
        }


    }
}