using System;
using System.Collections.Generic;
using System.Text;
using SharpNeatLib.NeatGenome;
using SharpNeatLib.Evolution;
using SharpNeatLib.NeuralNetwork;
using SharpNeatLib.NeatGenome.Xml;
using System.Xml;
using SharpNeatLib.CPPNs;
using SharpNeatLib.Experiments;
using System.Drawing;
using System.Xml.Serialization;
using System.Runtime.Serialization;
using System.IO;

namespace Engine
{
    //Basic Multi-Agent simulator experiment
    public class MultiAgentExperiment : SimulatorExperiment
    {
		public bool collisionPenalty;
        public bool multipleEnvironments;   //Evaluation over multiple environments?
        public bool evolveSubstrate;        //Not supported right now
        public bool modulatory;

        [XmlIgnore]
        public double elapsed;              //Elapsed simulation time

        public int numberRobots;
        public bool homogeneousTeam;
        public bool agentsVisible;          //Can the agents see each other
        public bool agentsCollide;          //Do the agents collide with each other

        public MultiAgentExperiment()
        {
			collisionPenalty = false;
            //heterogenous trams by default
            homogeneousTeam = false;

            scriptFile = "";
            numberRobots = 1;

            //how many range finders
            sensorDensity = 5;
            timestep = 0.16f;

            evaluationTime = 80;

            explanatoryText = "Multi-Agent Experiment";
            //Environment = new List<Environment>();

            //Create an empty environment
            environment = new Environment();
            robots = new List<Robot>();

            fitnessFunctionName = "SingleGoalPoint";
            behaviorCharacterizationName = "EndPointBC";
            robotModelName = "PackBotModel";

            initialized = false;
            running = false;

            //Determine the number of CPPN inputs and outputs
            if (homogeneousTeam)
                inputs = 4;
            else
                inputs = 5;

            if (adaptableANN)
            {
                if (modulatoryANN) outputs = 8; else outputs = 7;
            }
            else
                outputs = 2;

        }

        public override void initialize()
        {
            setupVariables();

            initialized = true;
        }

        protected void setupVariables()
        {
            substrateDescription = new SubstrateDescription(substrateDescriptionFilename);

            agentBrain = new AgentBrain(homogeneousTeam, numberRobots, substrateDescription, genome != null ? genome.Decode(null) : null, normalizeWeights, adaptableANN, modulatoryANN);
            
            //if(homogeneousTeam)
            //Console.WriteLine("inc:"+agentBrain.getBrain(0).InputNeuronCount);
            
            loadEnvironments(this);

            initializeRobots(agentBrain, environment, headingNoise, sensorNoise, effectorNoise);

            setFitnessFunction(fitnessFunctionName);
            setBehavioralCharacterization(behaviorCharacterizationName);


            if (gridCollision)
                collisionManager = new GridCollision();
            else
                collisionManager = new StandardCollision();
            collisionManager.Initialize(environment, this);

            elapsed = 0;
        }

        #region Environment

        public void loadEnvironment(String filename)
        {
            Console.WriteLine("loading environment...");
            environment = Environment.loadEnvironment(filename);
        }

        #endregion

        public override void draw(Graphics g, CoordinateFrame scale)
        {
            if (collisionManager is GridCollision)
            {
                GridCollision gc = (GridCollision)collisionManager;
                gc.draw(g, scale);
            }


            if (environment != null)
            {
                environment.draw(g, scale);

            }

            foreach (Robot robot in robots)
            {
                robot.draw(g, scale);
            }

            g.DrawString("Fitness: " + this.fitnessFunction.calculate(this, this.environment), new Font("Tahoma", 12), Brushes.Black, 10, 30);
            g.DrawString("Elapsed time: " + this.elapsed, new Font("Tahoma", 12), Brushes.Black, 10, 60);
        }


        public override void run()
        {
            runEnvironment(environment, (float)timestep, headingNoise, sensorNoise, effectorNoise);
        }

        //TODO this could be put into a Formation class or something that takes an environment as input
        public override void initializeRobots(AgentBrain agentBrain, Environment e, float headingNoise, float sensorNoise, float effectorNoise)
        {
            //Set up robots.
            //set up deltas for spacing the robots
            double dx, dy;
            if (overrideTeamFormation)
            {
                dx = Math.Cos(group_orientation / 180.0 * 3.14) * group_spacing;
                dy = Math.Sin(group_orientation / 180.0 * 3.14) * group_spacing;
            }
            else
            {
                dx = Math.Cos(e.group_orientation / 180.0 * 3.14) * e.robot_spacing;
                dy = Math.Sin(e.group_orientation / 180.0 * 3.14) * e.robot_spacing;
            }

            //record z-stack coordinates
            //float zstack = -1.0f;
            //float zdelta = 2.0f / (numberRobots - 1);

            robots.Clear();

            int heading = overrideTeamFormation ? robot_heading : e.robot_heading;

            //add robots in their formation according to environment
            for (int num = 0; num < numberRobots; num++)
            {
                Robot r = RobotModelFactory.getRobotModel(robotModelName);
                r.init(num, e.start_point.x + num * dx, e.start_point.y + num * dy,
                    heading / 180.0 * 3.14, agentBrain, e, sensorNoise, effectorNoise, headingNoise, (float)timestep);
				r.collisionPenalty=collisionPenalty;
                agentBrain.registerRobot(r);

                if (overrideDefaultSensorDensity)
                    r.populateSensors(sensorDensity);
                else
                    r.populateSensors();  //TODO populate sensors gets called multiple times. also in robot contructor
                if (agentBrain.zcoordinates == null) r.zstack = 0;
                else
                    r.zstack = agentBrain.zcoordinates[num];
                //zstack;
                robots.Add(r);

                //zstack += zdelta;
            }
            substrateDescription.setNeuronDensity(0, (uint)robots[0].sensors.Count, 1);
            agentBrain.updateInputDensity();
        }

        protected internal virtual bool runEnvironment(Environment e, float delta, float headingNoise, float sensorNoisde, float effectorNoise)
        {
            bool collide;
            elapsed += timestep;
            timeSteps++;
            collisionManager.SimulationStepCallback();

            for (int x = 0; x < robots.Count; x++)
            {


                robots[x].updateSensors(e, robots, collisionManager);

                //TODO: Apply input noise?  Maybe should apply it at the sensors (setup with addRobots)

                robots[x].doAction();


            }
            for (int x = 0; x < robots.Count; x++)
            {
                collide = collisionManager.RobotCollide(robots[x]);

                robots[x].collide_last = collide;
                if (collide)
                    robots[x].onCollision();
            }
            behaviorCharacterization.update(this);
            fitnessFunction.update(this, e);


            return false;
        }

        internal override double evaluateNetwork(INetwork network, out SharpNeatLib.BehaviorType behavior)
        {
            double fitness = 0;//SharpNeatLib.Utilities.random.NextDouble();
            NeatGenome tempGenome;
            INetwork controller;

            //TODO: Add novelty search stuff
            behavior = new SharpNeatLib.BehaviorType();
            foreach (Environment env in environmentList)
            {
                //decode genome and assign controllers to robots
               AgentBrain agentBrain = new AgentBrain(homogeneousTeam, numberRobots, substrateDescription, network, normalizeWeights, adaptableANN, modulatoryANN);
               
                double tempFit = 0;
                for (int evals = 0; evals < timesToRunEnvironments; evals++)
                {
                    // environment.reset();
                    //add robots to the environment
                    initializeRobots(agentBrain, env, headingNoise, sensorNoise, effectorNoise);
                    elapsed = 0;
                    timeSteps = 0;
                    while (elapsed < evaluationTime)
                    {
                        runEnvironment(env, (float)timestep, headingNoise, sensorNoise, effectorNoise);
                    }
                   
                    tempFit += fitnessFunction.calculate(this, env);
                    fitnessFunction.reset();
                    behavior.behaviorList = behaviorCharacterization.calculate(this);
                    //for(int i=0;i<behavior.behaviorList.Count;i++)
                    //		Console.Write(behavior.behaviorList[i]+" ");
                    //		Console.WriteLine();
                    //tempFit +=  EngineUtilities.euclideanDistance(robots[0].location, e.start_point);
                }
                fitness += tempFit / timesToRunEnvironments;
            }
            return fitness / environmentList.Count;
        }

        public override String defaultSeedGenome()
        {
            if (homogeneousTeam) return
                "seedGenome2x.xml";
            else return
                "seedGenomeHomo.xml";
        }

        #region IXmlSerializable Members

        public System.Xml.Schema.XmlSchema GetSchema()
        {
            throw new NotImplementedException();
        }

        #endregion
    }
}
