using System;
using System.Collections.Generic;
using System.Text;
using SharpNeatLib.Evolution;
using SharpNeatLib.NeuralNetwork;
using SharpNeatLib.CPPNs;
using SharpNeatLib.NeatGenome;
using SharpNeatLib.Experiments;
using System.Drawing;
using System.Xml;
using SharpNeatLib.NeatGenome.Xml;
using System.Diagnostics;
using System.Xml.Serialization;
using System.IO;

namespace Engine
{
   public abstract class SimulatorExperiment
    {
        [XmlIgnore]
        public AgentBrain agentBrain;
        [XmlIgnore]
        public SubstrateDescription substrateDescription;
        [XmlIgnore]
        public CollisionManager collisionManager;
        public bool gridCollision = false;

        public string substrateDescriptionFilename;

       //Is the ANN adatable
        public bool adaptableANN        = false;
       //Use neuromodulation
        public bool modulatoryANN       = false;
       //If true normalizes the weights of the ANN
        public bool normalizeWeights    = true;

        public int inputs, outputs;
        public int populationSize = 150;

        //The different types of noise
        public float headingNoise   = 0.0f;
        public float effectorNoise  = 0.0f;
        public float sensorNoise    = 0.0f;

       //How many times should the environment be run
        public int timesToRunEnvironments = 1;
        public int evaluationTime;

        [XmlIgnore]
        public bool initialized;

        public double timestep;
        public string explanatoryText;
        public string scriptFile;
        public String robotModelName;
        public string fitnessFunctionName;
		public string behaviorCharacterizationName;

        [XmlIgnore]
        public IFitnessFunction fitnessFunction;

        [XmlIgnore]
        public IBehaviorCharacterization behaviorCharacterization;

        //The current environment
        [XmlIgnore]
        public Environment environment;
        [XmlIgnore]
        public List<Environment> environmentList = new List<Environment>();
        [XmlIgnore]
        public int timeSteps = 0;

        public string environmentName;

        [XmlIgnore]
        public NeatGenome bestGenomeSoFar;

        //Robots
        [XmlIgnore]
        public List<Robot> robots = new List<Robot>();

        public bool overrideDefaultSensorDensity = false;
        public int sensorDensity;

        public bool overrideTeamFormation = false;
        public float group_orientation;
        public float group_spacing;
        public int robot_heading;

        public bool useScript = false;  
        [XmlIgnore] 
        public bool running;        //is the experiment running?

        ////genome
        [XmlIgnore]
        public NeatGenome genome;

        public void loadGenome(String filename)
        {
            XmlDocument doc = new XmlDocument();
            doc.Load(filename);
            genome = XmlNeatGenomeReaderStatic.Read(doc);
            bestGenomeSoFar = genome;
		    //TODO maybe include	initialize(); 
        }

		public void setBehavioralCharacterization(string behaviorName)
		{
			behaviorCharacterization = BehaviorCharacterizationFactory.getBehaviorCharacterization(behaviorName);
			behaviorCharacterization.reset();
		}

        public void setFitnessFunction(string fitnessFunctionName)
        {
            fitnessFunction = FitnessFunctionFactory.getFitnessFunction(fitnessFunctionName);
            fitnessFunction.reset();
        }

        public void loadEnvironment(String environment_name)
        {
            environmentName = environment_name;
			environment=Environment.loadEnvironment(environment_name);
        }

        public void clearEnvironment()
        {
            environment.reset();
        }

        internal abstract double evaluateNetwork(INetwork network, out SharpNeatLib.BehaviorType behavior);

        public void resetEnvironment()
        {
            environment.reset();
        }
        public virtual String defaultSeedGenome()
        {
            return null;
        }

        #region Virtual Methods
        public virtual void run() { }
        
        public virtual void draw(Graphics g, CoordinateFrame scale) { }

        public virtual void initialize() {}
        #endregion

        #region XML Serialization
        //Save the environment to XML
        public virtual void save(string name)
        {
            System.Xml.Serialization.XmlSerializer x = new System.Xml.Serialization.XmlSerializer(this.GetType());
            TextWriter outfile = new StreamWriter(name);
            x.Serialize(outfile, this);
            outfile.Close();
        }
        /*
        //Loads a environment from an XML file and initializes it
        public SimulatorExperiment load(string name)
        {
            System.Xml.Serialization.XmlSerializer x = new System.Xml.Serialization.XmlSerializer(typeof(SimulatorExperiment));
            TextReader infile = new StreamReader(name);
            SimulatorExperiment e = (SimulatorExperiment)x.Deserialize(infile);
            infile.Close();

            e.substrateDescription = new SubstrateDescription(e.substrateDescriptionFilename);

            e.fitnessFunction = FitnessFunctionFactory.getFitnessFunction(e.fitnessFunctionName);

            loadEnvironments(e);
            
            return e;
        }*/

        public static void loadEnvironments(SimulatorExperiment e)
        {
            e.environmentList.Clear();
            Engine.Environment scene = Engine.Environment.loadEnvironment(e.environmentName);
            e.environmentList.Add(scene);

            e.environment = scene;

            Console.Write("Looking for additional environments [" + scene.name + "] ... ");
            String filenamePrefix = scene.name.Substring(0, scene.name.Length - 4);
            int num = 1;
            String filename2 = filenamePrefix + num + ".xml";
            while (File.Exists(filename2))
            {
                Console.WriteLine("Found environment: " + filename2 + "\n");
                e.environmentList.Add(Engine.Environment.loadEnvironment(filename2));
                num++;
                filename2 = filenamePrefix + num + ".xml";
            }
			
            Console.WriteLine("Done");
            Console.WriteLine(e.environmentList.Count.ToString() + " environment(s) found.");
        }

        public abstract void initializeRobots(AgentBrain agentBrain, Environment e, float headingNoise, float sensorNoise, float effectorNoise);

        #endregion
    }


}
